/*************************************************************************************************************************************************
   ESP32cam test servo driver

   get command byte from the PC and act on it
     see pollSerial() for details

   if receives a byte from the Nano
     send it to PC over serial

 *************************************************************************************************************************************************/

#include "esp_camera.h"
#include "SPI.h"
#include "driver/rtc_io.h"
#include "SimpleILI9341sw.h"
#include "TBIB_Master.h"
#define CAMERA_MODEL_AI_THINKER
#include "camera_pins.h"

#define SERVOMIN  150 // This is the 'minimum' pulse length count (out of 4096)
#define SERVOMAX  600 // This is the 'maximum' pulse length count (out of 4096)
#define USMIN  600 // This is the rounded 'minimum' microsecond length based on the minimum pulse of 150
#define USMAX  2400 // This is the rounded 'maximum' microsecond length based on the maximum pulse of 600
#define SERVO_FREQ 50 // Analog servos run at ~50 Hz updates

const int Lmax = 150; // pixel maximum lightness - brighter and it's "white"
const int Lmin = 50; // pixel minimum lightness - darker and it's "black"
const int Smin = 30; // pixel minimum saturation - less and it's "gray"
const int yelmin = 20-10; const int yelmax = 55-10; // hue range for yellow
const int grnmin = yelmax+1; const int grnmax = 95-10; // hue range for green
const int cynmin = grnmax+1; const int cynmax = 135-10; // hue range for cyan
const int blumin = cynmax+1; const int blumax = 186-10; // hue range for blue
const int magmin = blumax+1; const int magmax = 230-10; // hue range for magenta 
// (anything else is "red")
const int blkmax = 100;
const int changemax = 1000;
const int wSample = 16; // size of sample window / 2
const int hSample = 16; // size of sample window / 2

#define SWAP_HI_LO(x) (((x << 8) & 0xFF00) | ((x >> 8) & 0xFF))

const int pinTFTCS = 12;

byte LearnedPosition[][8] = { // 0 means ignored
  { 78,108,123,  0,0,0,0,0},  // A Back
  {164, 72,123,  0,0,0,0,0},  // B Bck Dn
  {179, 99,123,  0,0,0,0,0},  // C Fwd Dn
  {113,130,176,  0,0,0,0,0},  // D R-Bin
  {104,145,212,  0,0,0,0,0},  // E Y-Bin
  {144,167,182,  0,0,0,0,0},  // F B-Bin
  {  0,  0,  0, 83,0,0,0,0},  // G Closed
  {  0,  0,  0,115,0,0,0,0}}; // H Open

byte LearnedSequence[][10] = {
  {0,1,2,6,0,4,7,0,255},  // N put in yellow bin
  {0,1,2,6,0,3,7,0,255},  // O put in red bin
  {0,1,2,6,0,5,7,0,255}}; // P put in blue bin

camera_fb_t * fbSend = NULL; // buffer for picture data

byte DetectMode = 2; // 0:Off; 1:Detect; 2:Detect and Act
byte nDetect;
uint16_t prevCol;
bool bLED = false;
bool bAtTargets = false;
byte prevLearnedPos = 255;

//-----------------------------------------------------------
// esp_err_t initialise_camera
//-----------------------------------------------------------
esp_err_t initialise_camera() {
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;

  config.pixel_format = PIXFORMAT_RGB565;
  //  config.pixel_format = PIXFORMAT_GRAYSCALE;
  config.frame_size = FRAMESIZE_QVGA;
  //  config.frame_size = FRAMESIZE_QQVGA;
  config.jpeg_quality = 10;
  config.fb_count = 1;

  //initialize camera
  return esp_camera_init(&config);
}

//-----------------------------------------------------------
// take_photo
//-----------------------------------------------------------
void take_photo() {
  if (bLED) {
    pinMode(4, OUTPUT);              //LED flash
    digitalWrite(4, HIGH);            //flash LED ON
//    delay(1);
  }
  //take photo

  if (fbSend != NULL)
    esp_camera_fb_return(fbSend);

  fbSend = esp_camera_fb_get();
  if (!fbSend)
  {
    Serial.println("photo failed");
    while (1);  //stop
  }
  digitalWrite(4, LOW);            //flash LED OFF
}

//-----------------------------------------------------------
// display_photo
//-----------------------------------------------------------
void display_photo() {
  while (!PollTBIB()) {}; // only use SPI when both clocks are low

  pinMode(pinTFTCS, OUTPUT);
  pinMode(pinMISOa, OUTPUT);
  pinMode(pinSCLK, OUTPUT);
  pinMode(pinMOSIa, OUTPUT);
  digitalWrite(pinTFTCS, HIGH);
  digitalWrite(pinMISOa, HIGH);
  digitalWrite(pinSCLK, HIGH);
  digitalWrite(pinMOSIa, HIGH);
  delay(1);

  if (fbSend->format == PIXFORMAT_RGB565)
    DrawBitmapArrayReverse(0, 0, fbSend->width, fbSend->height, (unsigned byte *)(fbSend->buf));
  //    DrawBitmapArrayReverse(0, 0, fbSend->width, fbSend->height, (unsigned short *)(fbSend->buf));
  //    DrawBitmapArray(0, 0, fbSend->width, fbSend->height, (unsigned short *)(fbSend->buf));

  if (fbSend->format == PIXFORMAT_GRAYSCALE)
    DrawBitmapArrayByte(0, 0, fbSend->width, fbSend->height, (unsigned byte *)(fbSend->buf));

  DrawFrame(
    fbSend->width/2-wSample-1,
    fbSend->height/2-hSample-1,
    wSample*2+1,
    hSample*2+1,
    TFT_RED);

  InitTBIBpins();
}

//-----------------------------------------------------------
// camera_settings
//-----------------------------------------------------------
void camera_settings() {
  sensor_t * s = esp_camera_sensor_get();

  s->set_gain_ctrl(s, 0); // auto gain off (1 or 0)
  s->set_exposure_ctrl(s, 1); // auto exposure  (1 or 0)
  s->set_agc_gain(s, 0); // set gain manually (0 - 30)
  s->set_aec_value(s, 300); // set exposure manually (0-1200)
  s->set_contrast(s, -2); // (-2 to 2)

  //  s->set_brightness(s, -2);     // -2 to 2

  // s->set_brightness(s, 0); // (-2 to 2) - set brightness
  // s->set_awb_gain(s, 0); // Auto White Balance?
  // s->set_lenc(s, 0); // lens correction? (1 or 0)
  // s->set_raw_gma(s, 1); // (1 or 0)?
  // s->set_quality(s, 10); // (0 - 63)
  // s->set_whitebal(s, 1); // white balance
  // s->set_wb_mode(s, 1); // white balance mode (0 to 4)
  // s->set_aec2(s, 0); // automatic exposure sensor? (0 or 1)
  // s->set_aec_value(s, 0); // automatic exposure correction? (0-1200)
  // s->set_saturation(s, 0); // (-2 to 2)
  // s->set_hmirror(s, 0); // (0 or 1) flip horizontally
  // s->set_gainceiling(s, GAINCEILING_32X); // Image gain (GAINCEILING_x2, x4, x8, x16, x32, x64 or x128)
  // s->set_sharpness(s, 0); // (-2 to 2)
  // s->set_colorbar(s, 0); // (0 or 1) - testcard
  // s->set_special_effect(s, 0);
  // s->set_ae_level(s, 0); // auto exposure levels (-2 to 2)
  // s->set_bpc(s, 0); // black pixel correction
  // s->set_wpc(s, 0); // white pixel correction
  // s->set_dcw(s, 1); // downsize enable? (1 or 0)?
}

//-----------------------------------------------------------
// send_photo_to_PC()
//-----------------------------------------------------------
void send_photo_to_PC() {
  uint8_t *pSend = NULL; // pointer used to send picture data to PC
  uint32_t i; // counter used to send picture data to PC

  pSend = fbSend->buf;
  Serial.print("\n\n\nImage ");
  Serial.print(fbSend->width); Serial.print(" ");
  Serial.print(fbSend->height); Serial.print(" ");
  Serial.print(fbSend->len); Serial.print(" ");
  Serial.print(fbSend->format); Serial.print(":");

  for (i = 0; i < fbSend->len; i++) {
    Serial.write(*pSend);
    pSend++;
  }
}

//-----------------------------------------------------------
// recvByte()
//  wait for a byte from the PC
//-----------------------------------------------------------
byte recvByte() {
  while (! Serial.available()) ;
  byte b = Serial.read();
  return b;
}

//-----------------------------------------------------------
// SendLearnedPos()
//    move servos to learned position
//-----------------------------------------------------------
void SendLearnedPos(byte i) {
  if (prevLearnedPos == i) {
    bAtTargets = true;    
    return;
  }
  prevLearnedPos = i;
  
  byte *pos = LearnedPosition[i];
  for (byte servo=0; servo < 8; servo++) 
  if (pos[servo])
  {
    Serial.print("set pos "); Serial.print(servo); Serial.print(" "); Serial.println(pos[servo]); 
    StartMsgToSlave(msgSetServo);
    SendByteToSlave(servo);
    SendByteToSlave(pos[servo]);
    EndMsgToSlave();
  }
}
        
//-----------------------------------------------------------
// SendLearnedSeq()
//    move servos in a learned sequence
//-----------------------------------------------------------
void SendLearnedSeq(byte i) {
  byte *seq = LearnedSequence[i];
  for (byte pos=0; pos < 8; pos++) 
  {
    if (seq[pos] == 255)
      return;
    unsigned long t = millis();
    bAtTargets = false;
    SendLearnedPos(seq[pos]);
    while ((millis() - t < 3000) && (! bAtTargets))
      PollTBIB();
  }
}
        
//-----------------------------------------------------------
// SerialPrintlnColor()
//-----------------------------------------------------------
void SerialPrintlnColor(uint16_t col) {
  switch (col) {
    case TFT_WHITE:    Serial.println("WHITE"); break;
    case TFT_BLACK:    Serial.println("BLACK"); break;
    case TFT_DARKGREY: Serial.println("GREY"); break;
    case TFT_YELLOW:   Serial.println("YELLOW"); break;
    case TFT_GREEN:    Serial.println("GREEN"); break;
    case TFT_CYAN:     Serial.println("CYAN"); break;
    case TFT_BLUE:     Serial.println("BLUE"); break;
    case TFT_MAGENTA:  Serial.println("MAGENTA"); break;
    case TFT_RED:      Serial.println("RED"); break;
  }          
}
        
//-----------------------------------------------------------
// pollSerial()
//   get command byte from the PC and act on it
//     'p': take a photo and send to PC over serial
//     'q': take a photo and display on ILI9341
//     'A'..'M': move servos to learned position
//     'N'..'Z': move servos in a learned sequence
//     'g': send QueryPWMs message to Nano
//     '0'..'7': send SetServo position message to Nano
//     'd': detect mode on/off
//     's': detect-and-sort mode on/off
//     'l': LED on/off
//     'c': measure pixel in centre of screen, RGB, HSL, etc.
//     else send it to Nano
//-----------------------------------------------------------
void pollSerial() {
  if (Serial.available() > 0) {
    byte b = recvByte();
    
    switch (b) {
      case 'd': case 's': break;
      default: DetectMode = 0;
    }
    
    switch (b) {
      case 'p':
        Serial.println("Send Photo");
        take_photo();
        send_photo_to_PC();
        break;
      case 'q':
        {
          Serial.println("Display Photo");
          take_photo();
          display_photo();
        }
        break;
      case 'g':
        Serial.println("Query PWM");
        SendMsgToSlave(msgPWMs);
        break;

      case 'd':
        DetectMode = DetectMode == 0;
        Serial.print("Detect mode ");
        Serial.println(DetectMode);
        break;

      case 's':
        DetectMode = 2*(DetectMode == 0);
        Serial.print("Detect mode ");
        Serial.println(DetectMode);
        nDetect = 0;
        prevCol = 0;
        break;

      case 'l':
        bLED = ! bLED;
        Serial.print("bLED ");
        Serial.println(bLED);
        break;

      case 'c':
        {
          Serial.println("Detect color");
          uint16_t col = pixel(160,120);
          Serial.print("col "); Serial.println(col,HEX);
          Serial.print("rgb "); Serial.print(unpackR(col)); Serial.print(" "); Serial.print(unpackG(col)); Serial.print(" "); Serial.println(unpackB(col));
          byte exH,exS,exL;
          RGBtoHSL(unpackR(col), unpackG(col), unpackB(col), &exH,&exS,&exL);
          Serial.print("hsl "); Serial.print(exH); Serial.print(" "); Serial.print(exS); Serial.print(" "); Serial.println(exL);
          int32_t total;
          SerialPrintlnColor(DetectColor(160-wSample,120-hSample,160+wSample,120+hSample, &total));
        }
        break;

      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
        {
          byte pos = recvByte();
          Serial.print("set pos ");
          Serial.print((char)b);
          Serial.print(" = ");
          Serial.println(pos);
          StartMsgToSlave(msgSetServo);
          SendByteToSlave(b - '0');
          SendByteToSlave(pos);
          EndMsgToSlave();
        }
        break;
        
      case 'A':
      case 'B':
      case 'C':
      case 'D':
      case 'E':
      case 'F':
      case 'G':
      case 'H':
        SendLearnedPos(b-'A');
        break;
        
      case 'N':
      case 'O':
      case 'P':
        Serial.print("SEQUENCE ");
        Serial.println((char)b);
        SendLearnedSeq(b-'N');
        break;

//      default:
//        SendByteToSlave(b);
//        Serial.print("Tx from Serial ");
//        Serial.println(b);
//        break;
    }
  }
}

//-----------------------------------------------------------
// recvTBIBmsg
//     msgInitialise: nano has Initialised itself
//     msgAtTargets: the servos have reached their target positions
//     msgPWMs: Nano reports the current servo positions
//-----------------------------------------------------------
void recvTBIBmsg(byte msg, byte *data) {
  Serial.print("Rx Msg "); Serial.println(msg);
  switch (msg) {
    case msgInitialise:
      Serial.println("Initialise");
      break;
    case msgAtTargets:
      Serial.println("AtTargets");
      bAtTargets = true;
      break;
    case msgPWMs:
      Serial.println("PWMs");
      for (int i = 0; i <= 7; i++) 
        Serial.println(data[i * 2] + (data[i * 2 + 1] << 8));
      break;
  }
}

//-----------------------------------------------------------
// RGBtoHSL
//     convert RGB to HSL
//-----------------------------------------------------------
void RGBtoHSL(uint8_t R, uint8_t G, uint8_t B, uint8_t *H, uint8_t *S, uint8_t *L) {
  int32_t D, Cmax, Cmin;
  int32_t HL;

  R = R*240 / 255;
  G = G*240 / 255;
  B = B*240 / 255;

  Cmax = max(R, max(G, B));
  Cmin = min(R, min(G, B));
  *L = constrain((Cmax + Cmin) / 2,0,255);

  if (Cmax == Cmin)
  {
    *H = 0;
    *S = 0;
  } else {
    D = (Cmax - Cmin) * 240;
    if (*L <= 120)
      *S = constrain(D / (Cmax + Cmin),0,240);
    else
      *S = constrain(D / (240 * 2 - Cmax - Cmin),0,240);

    D = D * 6;
    if (R == Cmax)
      HL = (G - B) * 240 * 240 / D;
    else if (G == Cmax)
      HL = 240 * 2 / 6 + (B - R) * 240 * 240 / D;
    else
      HL = 240 * 4 / 6 + (R - G) * 240 * 240 / D;

    if (HL < 0)
      HL = HL + 240;
    *H = HL;
  }
}

//-----------------------------------------------------------
// pixel
//     get a pixel from the camera buffer
//-----------------------------------------------------------

uint16_t pixel(int x, int y) {
  if (fbSend->format == PIXFORMAT_RGB565) {
    uint16_t i = ((uint16_t *)(fbSend->buf))[y*fbSend->width+x];
    return SWAP_HI_LO(i);
  } else 
    return ((uint8_t *)(fbSend->buf))[y*fbSend->width+x];
}

//-----------------------------------------------------------
// DetectColor
//   what is the "main" color inside a rectangle of the camera buffer
//   determine the "color" of each pixel
//   vote on which is the most likely
//     black, white and gray votes count for less because they're "background"
//     yellow votes count double because it's harder to detect
//-----------------------------------------------------------
uint16_t DetectColor(int x1, int y1, int x2, int y2, int32_t *total) {
  int n,w;
  byte exH,exS,exL;
  int nwhite,nblack,ngray,nyellow,ngreen,ncyan,nblue,nmagenta,nred;
  uint32_t col;

  nwhite = nblack = ngray = nyellow = ngreen = ncyan = nblue = nmagenta = nred = *total =0;
  for (int ax = min(x1,x2); ax <= max(x1,x2); ax++)
  for (int ay = min(y1,y2); ay <= max(y1,y2); ay++)
  {
    col = pixel(ax,ay);
    RGBtoHSL(unpackR(col), unpackG(col), unpackB(col), &exH,&exS,&exL);
    *total += exL;
    if (max(unpackR(col), max(unpackG(col), unpackB(col))) < blkmax)    nblack++; else
    if (exL > Lmax)    nwhite++; else
    if (exL < Lmin)    nblack++; else
    if (exS < Smin)    ngray++; else
    if (exH <  yelmin) nred++; else
    if (exH <= yelmax) nyellow++; else
    if (exH <= grnmax) ngreen++; else
    if (exH <= cynmax) ncyan++; else
    if (exH <= blumax) nblue++; else
    if (exH <= magmax) nmagenta++; else
                       nred++;
  }

  w = 0;
  col = TFT_BLACK;
  nwhite   = nwhite  / 6; if (nwhite > w)   { w = nwhite;    col = TFT_WHITE; }
  nblack   = nblack  / 6; if (nblack > w)   { w = nblack;    col = TFT_BLACK; }
  ngray    = ngray   / 6; if (ngray > w)    { w = ngray;     col = TFT_DARKGREY; }
  nyellow  = nyellow * 2; if (nyellow > w)  { w = nyellow;   col = TFT_YELLOW; }
  ngreen   = ngreen  ;    if (ngreen > w)   { w = ngreen;    col = TFT_GREEN; }
  ncyan    = ncyan   ;    if (ncyan > w)    { w = ncyan;     col = TFT_CYAN; }
  nblue    = nblue   ;    if (nblue > w)    { w = nblue;     col = TFT_BLUE; }
  nmagenta = nmagenta;    if (nmagenta > w) { w = nmagenta;  col = TFT_MAGENTA; }
  nred     = nred    ;    if (nred > w)     { w = nred;      col = TFT_RED; }

  return col;
}

//-----------------------------------------------------------
// ReportColor
//     take a photo
//     detect the color
//     send to PC as a string
//-----------------------------------------------------------
void ReportColor() {
  take_photo();
  display_photo();
  int32_t total;
  static int32_t prev_total = 0;
  
  uint16_t col = DetectColor(160-wSample,120-hSample,160+wSample,120+hSample, &total);
//  Serial.println(abs(total - prev_total));
  
  if (abs(total - prev_total) > changemax) {
    Serial.println("Changing"); 
    nDetect = 0;
  } else {
    if (col == prevCol)
      nDetect++;
    prevCol = col;
    
    switch (col) {
      case TFT_WHITE: case TFT_BLACK: case TFT_DARKGREY: 
        nDetect = 0; 
        break;
    }          
    
    SerialPrintlnColor(col);

    if ((nDetect == 5) && (DetectMode == 2)) {
      SortBlock(col);
      nDetect = 0;
    }
  }
  
  prev_total = total;
}

//-----------------------------------------------------------
// SortBlock
//   move block to correct bin
//-----------------------------------------------------------
void SortBlock(uint16_t col) {
//  DetectMode = 0;
  switch (col) {
    case TFT_YELLOW:   SendLearnedSeq(0); break;
    case TFT_RED:      SendLearnedSeq(1); break;
    case TFT_BLUE:     SendLearnedSeq(2); break;
  }
}

//-----------------------------------------------------------
// setup
//-----------------------------------------------------------
void setup()
{
  Serial.begin(115200);
  Serial.println();
  Serial.println("Booting...");

  //  pinMode(4, OUTPUT);              //LED flash
  //  digitalWrite(4, HIGH);
  pinMode(4, INPUT);              //GPIO for LED flash
  digitalWrite(4, LOW);
  rtc_gpio_hold_dis(GPIO_NUM_4);  //disable pin hold if it was enabled before sleeping

  /**** this code magically makes the camera initialisation work reliably ****/
  /**** it turns the camera off and on ****/
  digitalWrite(PWDN_GPIO_NUM, LOW);
  delay(10);
  digitalWrite(PWDN_GPIO_NUM, HIGH);
  delay(10);

  esp_err_t err = initialise_camera();

  if (err != ESP_OK)
  {
    Serial.printf("Camera init failed: 0x%x", err);
    while (1);  //stop
  }

  delay(100);

  camera_settings();
  take_photo();
  Serial.println("camera init OK");

  ILI9341Begin(320, 240, ILI9341_Rotation3);

  //  DrawBox(0  ,  0,  106,  120, TFT_BLUE   );
  //  DrawBox(106,  0,  106,  120, TFT_GREEN  );
  //  DrawBox(212,  0,  106,  120, TFT_CYAN   );
  //  DrawBox(0  ,  120,  106,  120, TFT_RED    );
  //  DrawBox(106,  120,  106,  120, TFT_MAGENTA);
  //  DrawBox(212,  120,  106,  120, TFT_YELLOW );

  Serial.println("ILI9341 init OK");

  InitTBIBpins();

  Serial.println("Initialised");
}

//-----------------------------------------------------------
// main loop
//-----------------------------------------------------------
void loop() {
  pollSerial();
  PollTBIB();

  if (DetectMode)
    ReportColor();
}
